@@ -1,6 +1,9 @@
+import os
 import pydash as __
 from pprint import pformat
 import time
+import logging
+import useful.logs
 import json
 from typing import Callable, Dict
 from json.decoder import JSONDecodeError
@@ -11,7 +14,7 @@ from enum import Enum
 from pygqlc import GraphQLClient
 from singleton_decorator import singleton
 from .Notifications import NotificationBehaviour
-from .Logging import log, LogLevel, LogStyle
+from .Logging import log, json_log, LogLevel, LogStyle
 from . import uploaders
 from . import queries
 from . import mutations
@@ -19,6 +22,22 @@ from . import subscriptions
 from .dateHelpers import getUtcDate
 from .croniterHelpers import get_croniter
 from .redis import WorkerRedis
+import traceback
+
+import functools
+
+# helps to know if a function has been called at least once
+def trackcalls(func):
+    @functools.wraps(func)
+    def wrapper(*args, **kwargs):
+        wrapper.has_been_called = True
+        wrapper.call_args = args
+        wrapper.call_kwargs = kwargs
+        return func(*args, **kwargs)
+    wrapper.has_been_called = False
+    wrapper.call_args = []
+    wrapper.call_kwargs = {}
+    return wrapper
 
 
 class QueueType(Enum):
@@ -171,7 +190,11 @@ class ValiotWorker():
         self.redis = WorkerRedis()
         self.continuousJobs = []
         # logging attributes
-        self.loggingStyle = LogStyle.PREFIX_FIRST_LINE
+        current_env = os.environ.get('ENV')
+        if current_env == 'dev':
+            self.setLoggingStyle(LogStyle.PREFIX_FIRST_LINE)
+        else:
+            self.setLoggingStyle(LogStyle.JSON)
         self.stopLogging = False
         self.loggingProcess = None
         self.log = None
@@ -562,6 +585,7 @@ class ValiotWorker():
             kwargs = job_dict
         # try to run job
 
+        @trackcalls
         def update_this_job(variables):
             # * funcion de utilidad para actualizar job sin necesidad de conocer su ID
             status = variables.get('status')
@@ -579,6 +603,8 @@ class ValiotWorker():
                 'queue': __.get(job, 'queue.name'),
                 'context': json.dumps(context)
             })
+        
+        log_from_job = self.getProcessLoggerFunction(queue=job['queue'], job=job)
         src_status_conn, dst_status_conn = Pipe()
 
         def get_job_status():
@@ -597,8 +623,7 @@ class ValiotWorker():
         kwargs['get_job_status'] = get_job_status
         # acceso directo a atributos del queue al que pertenece el job
         kwargs['queue'] = job['queue']
-        kwargs['log'] = self.getProcessLoggerFunction(
-            queue=job['queue'], job=job)
+        kwargs['log'] = log_from_job
         kwargs['context'] = __.get(job, 'queue.context', dict())
         kwargs['update_context'] = update_this_context
         kwargs['kwargs'] = {**kwargs}
@@ -614,17 +639,32 @@ class ValiotWorker():
         queue_success = __.get(data, 'updateQueue.successful')
         if job_success and queue_success:
             def safe_fn(**kwargs):
+                output = None
+                errors = []
                 try:
-                    job_fn(**kwargs)
+                    output = job_fn(**kwargs)
                 except TypeError as e:
-                    self.log(LogLevel.ERROR,
-                             f'Error starting job from queue {fn_name}:\n{str(e)}')
-                    self.log(LogLevel.ERROR,
-                             f'Did you forget to ignore additional params in the job definition?:')
+                    self.log(
+                        LogLevel.ERROR, f'Error starting job from queue {kwargs["queue"]["name"]}:\n{str(e)}')
+                    self.log(
+                        LogLevel.ERROR, f'Did you forget to ignore additional params in the job definition?:')
                     self.log(LogLevel.ERROR, f'@vw.job(...)')
                     self.log(
                         LogLevel.ERROR, f'def my_job(my_param, my_param2, **_) # <== can be **kwargs or **_')
                     self.log(LogLevel.ERROR, f'\t...')
+                except Exception as e:
+                    error_info = f"{str(e)}\nstack trace:\n{traceback.format_exc()}" if self.loggingStyle != LogStyle.JSON else ""
+                    error = {
+                        "message": str(e),
+                        "stack": traceback.format_exc()
+                    }
+                    errors.append(error)
+                    self.log(LogLevel.ERROR, f"job {job['id']} crashed unexpectedly.\n{error_info}", extra={
+                        "error": error
+                    })
+                finally:
+                    self.finish_job_gracefully(
+                        job=job, job_kwargs=kwargs, output=output, errors=errors)
             p = self.context.Process(target=safe_fn, kwargs=(kwargs))
             self.log(LogLevel.INFO, f'started job with ID {job["id"]}')
             p.start()
@@ -837,6 +877,41 @@ class ValiotWorker():
                 finished.append(job_id)
         for f in finished:
             self.runningJobs.pop(f, None)  # remove any finished jobs
+    
+    def finish_job_gracefully(self, job, job_kwargs, output, errors = []):
+        if len(errors) > 0:
+            self.log(LogLevel.ERROR,
+                f"terminating job with ID {job['id']} with errors...")
+            return uploaders.update_job({
+                'id': job['id'],
+                'status': JobStatus.ERROR.value,
+                'output': json.dumps({'data': output, 'errors': errors})
+            })
+        elif output is None:
+            self.log(LogLevel.INFO, f"finished job with ID {job['id']} with implicit exit")
+        elif type(output) == str:
+            self.log(LogLevel.INFO, f"finished job with ID {job['id']} with message: \n{output}")
+        elif type(output) == dict:
+            self.log(
+                LogLevel.INFO, f"finished job with ID {job['id']} with data: \n{output}")
+
+        # if the job did call "update_job", we gotta make sure we don't override the last info sent explicitly:
+        update_job_fn = job_kwargs["update_job"]
+        if update_job_fn.has_been_called:
+            maybe_status = update_job_fn.call_args[0].get("status") if len(update_job_fn.call_args) > 0 else None
+            if maybe_status is not None:
+                last_status = JobStatus(maybe_status)
+                if (last_status in [JobStatus.FINISHED, JobStatus.ERROR, JobStatus.ABORTED, JobStatus.PAUSED]):
+                    # job had already been finished explicitly
+                    self.log(LogLevel.INFO, f"job {job['id']} terminated itself correctly")
+                    return
+                # if it was been called, but had a status of "RUNNING", it still requires an implicit termination:
+        return uploaders.update_job({
+            'id': job['id'], 
+            'status': JobStatus.FINISHED.value,
+            'output': json.dumps({'data': output, 'errors': []})
+            })
+        
 
     def unlockFinishedJobs(self, force=False):
         """This function updates all finished jobs to unlock them..
@@ -995,6 +1070,7 @@ class ValiotWorker():
             return True
         job = __.get(response, 'job.result')
 
+        @trackcalls
         def update_this_job(variables):
             # * funcion de utilidad para actualizar job sin necesidad de conocer su ID
             status = variables.get('status')
@@ -1034,8 +1110,10 @@ class ValiotWorker():
         kwargs['kwargs'] = {**kwargs}
 
         def safe_fn(**kwargs):
+            output = None
+            errors = []
             try:
-                self.eventQueues[queue_name]['function'](**kwargs)
+                output = self.eventQueues[queue_name]['function'](**kwargs)
             except TypeError as e:
                 self.log(
                     LogLevel.ERROR, f'Error starting job from queue {queue_name}:\n{str(e)}')
@@ -1045,6 +1123,18 @@ class ValiotWorker():
                 self.log(
                     LogLevel.ERROR, f'def my_job(my_param, my_param2, **_) # <== can be **kwargs or **_')
                 self.log(LogLevel.ERROR, f'\t...')
+            except Exception as e:
+                error_info = f"{str(e)}\nstack trace:\n{traceback.format_exc()}" if self.loggingStyle != LogStyle.JSON else ""
+                error = {
+                    "message": str(e),
+                    "stack": traceback.format_exc()
+                }
+                errors.append(error)
+                self.log(LogLevel.ERROR, f"job {job['id']} crashed unexpectedly.\n{error_info}", extra={
+                    "error": error
+                })
+            finally:
+                self.finish_job_gracefully(job=job, job_kwargs=kwargs, output=output, errors=errors)
         p = self.context.Process(target=safe_fn, kwargs=(kwargs))
         p.start()
         self.runningJobs[job['id']] = {
@@ -1193,15 +1283,31 @@ class ValiotWorker():
         job : int, optional
           ID for the job sending this message, by default 0
         """
-        job_zero_padding = len(str(__.get(self.worker, 'jobs.0.id')))
-        # no need to pass as argument, will always be the same
+        job_zero_padding = len(str(__.get(self.worker, 'jobs.0.id')))        
+        log_queue_key = __.get(queue, 'name', 'main')
+        log_queue = self.logQueues[log_queue_key]
+        # to log this in a single line, we prefer to keep it small
         W = str(self.worker["id"]).zfill(2)
         Q = str(__.get(queue, 'id', 0)).zfill(3)
         J = str(__.get(job, 'id', 0)).zfill(job_zero_padding)
-        log_queue_key = __.get(queue, 'name', 'main')
-        log_queue = self.logQueues[log_queue_key]
-        if self.loggingStyle == LogStyle.PREFIX_ALL_LINES:
-            def sendLogToQueue(level, message):
+        if self.loggingStyle == LogStyle.JSON:
+            # as we log this in JSON format, it can be human-readable
+            W = self.worker["code"]
+            Q = __.get(queue, 'code', "WORKER_PROCESS")
+            J = __.get(job, 'id', 0)
+            def sendLogToQueue(level, message, extra = {}):
+                D = getUtcDate(datetime.now()).strftime("%Y-%m-%dT%H:%M:%SZ")
+                log_queue.put({
+                    'level': level,
+                    'date': D,
+                    'worker': W,
+                    'queue': Q,
+                    'job': J,
+                    'content': message,
+                    'extra': extra
+                })
+        elif self.loggingStyle == LogStyle.PREFIX_ALL_LINES:
+            def sendLogToQueue(level, message, extra = {}):
                 D = getUtcDate(datetime.now()).strftime("%Y-%m-%dT%H:%M:%SZ")
                 msgs = message.split('\n')
                 for msg in msgs:
@@ -1210,7 +1316,7 @@ class ValiotWorker():
                         'content': f'[D={D}/W={W}/Q={Q}/J={J}] - {msg}'
                     })
         else:
-            def sendLogToQueue(level, message):
+            def sendLogToQueue(level, message, extra = {}):
                 D = getUtcDate(datetime.now()).strftime("%Y-%m-%dT%H:%M:%SZ")
                 log_queue.put({
                     'level': level,
@@ -1235,7 +1341,16 @@ class ValiotWorker():
                     for _, log_queue in self.logQueues.items():
                         while not log_queue.empty():
                             msg = log_queue.get()
-                            log(msg['level'], msg['content'])
+                            extra = {
+                                'job_date': msg['date'],
+                                'worker': msg['worker'],
+                                'queue': msg['queue'],
+                                'job': msg['job'],
+                                **msg['extra']
+                            } if self.loggingStyle == LogStyle.JSON else {}
+                            # pick the proper logging function to actually render the stream to stdin
+                            log_fn = json_log if self.loggingStyle == LogStyle.JSON else log
+                            log_fn(msg['level'], msg['content'], extra)
             except KeyboardInterrupt:
                 log(LogLevel.WARNING, 'Logging stopped due to keyboard interrupt')
             log(LogLevel.WARNING, 'Logging loop stopped')
